Package: icecdr
Type: Package
Title: Download Sea Ice Concentration Data from the NSIDC Climate Data
        Record
Version: 1.2.0
Authors@R: person("Elio", "Campitelli", , "eliocampitelli@gmail.com", role = c("cre", "aut", "cph"),
           comment = c(ORCID = "0000-0002-7742-9230"))
Description: Programmatic access to NSIDC's sea ice concentration CDR <https://nsidc.org/data/g02202> via ERDAPP server and Sea Ice index <https://nsidc.org/data/g02135>. Supports caching results and optional fixes for some inconsistencies of the raw files.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.3.0)
Suggests: here, ncdf4, rcdo, testthat (>= 3.0.0), vcr
Imports: cli, digest, glue, httr2, rlang
URL: https://github.com/eliocamp/icecdr,
        https://eliocamp.github.io/icecdr/
BugReports: https://github.com/eliocamp/icecdr/issues
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 05:16:33 UTC; user1
Author: Elio Campitelli [cre, aut, cph] (ORCID:
    <https://orcid.org/0000-0002-7742-9230>)
Maintainer: Elio Campitelli <eliocampitelli@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-08 05:31:46 UTC
Built: R 4.6.0; ; 2026-05-08 06:40:45 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cdr)
export(cdr_antarctic_daily)
export(cdr_antarctic_monthly)
export(cdr_arctic_daily)
export(cdr_arctic_monthly)
export(cdr_fix)
export(cdr_fix_grid)
export(cdr_fix_names)
export(sea_ice_index)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# icecdr

<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/eliocamp/icecdr/graph/badge.svg)](https://app.codecov.io/gh/eliocamp/icecdr)
[![R-CMD-check](https://github.com/eliocamp/icecdr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/eliocamp/icecdr/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The goal of icecdr is to download sea ice concentration data from the
NSIDC Climate Data Record.

## Installation

You can install the icecdr with:

``` r
install.packages("icecdr")
```

## Example

Use the `cdr_*` functions to download satellite-derived Antarctic or
Arctic sea ice concentration data at monthly or daily resolution. They
use version 5 by default but version 4 is also supported.

The basic usage downloads a NetCDF file in a temporary directory.

``` r
library(icecdr)

dates <- c("2020-01-01", "2023-01-01")
cdr_antarctic_monthly(dates, dir = "data")
#> Downloading data.
#> [1] "data/05c77e3b938053bdbed2137c61dc6052.nc"
```

By default, files are only downloaded if needed.

``` r
system.time(cdr_antarctic_monthly(dates, dir = "data"))
#> Returning existing file.
#>    user  system elapsed 
#>   0.019   0.000   0.020
```

There are four simple functions to download whole-domain data:

``` r
cdr_antarctic_daily()
cdr_antarctic_monthly()

cdr_arctic_daily()
cdr_arctic_monthly()
```

But the `cdr()` function exposes all arguments to download any crazy
subset

``` r
cdr(date_range = dates,
    # Data every 7 days
    date_stride = 7,
    resolution = "daily", 
    # Thin the grid by taking every other gridpoint
    xgrid_stride = 2,
    ygrid_stride = 2, 
    hemisphere = "north"
    )
```

The `cdr_fix()` function will fix the grid information to make it work
with [CDO](https://code.mpimet.mpg.de/projects/cdo) and standardise
variable names. This requires CDO installed in your system.

``` r
library(rcdo)
library(ggplot2)

extent_cdr <- cdr_antarctic_monthly(dates, dir = "data") |> 
  cdr_fix() |> 
  cdo_gtc(0.15) |>
  cdo_fldint() |> 
  cdo_execute() |> 
  metR::ReadNetCDF(c(extent = "aice")) 
#> Warning: Using CDO version 2.4.0 which is different from the version supported by this
#> version of rcdo (2.5.1).
#> ℹ This warning is displayed once per session.
#> Returning existing file.

extent_index <- sea_ice_index("south", "monthly", dir = "data") |> 
  data.table::fread() |>
  subset(time >= dates[1] & time <= dates[2])
#> Downloading data.

extent_cdr |>
  ggplot(aes(time, extent)) +
  geom_line(aes(colour = "cdr")) +
  geom_line(data = extent_index, aes(colour = "sea_ice_index")) +
  labs(y = "Sea ice extent")
```

<img src="man/figures/README-unnamed-chunk-3-1.png" alt="" width="100%" />

# nocov start
make_cdr <- function(definition) {
  force(definition)
  function(
    date_range,
    variables = "aice",
    version = 6,
    file = NULL,
    dir = tempdir(),
    use_cache = TRUE
  ) {
    cdr(
      date_range = date_range,
      variables = variables,
      hemisphere = definition$hemisphere,
      resolution = definition$resolution,
      version = version,
      file = file,
      dir = dir,
      use_cache = use_cache
    )
  }
}
# nocov end

#' Convenience functions
#'
#' These are convenience functions to download, daily or monthly
#' sea ice concentration data from the whole Antarctic or Arctic.
#' For a more complete low level function, see [cdr()].
#'
#' @inheritParams cdr
#'
#' @returns A string vector with the location of the downloaded files
#'
#' @examples
#' \dontrun{
#' cdr_antarctic_monthly(c("2022-01", "2022-01"))
#' }
#'
#'
#' @export
#' @rdname convenience
cdr_antarctic_monthly <- make_cdr(list(
  hemisphere = "south",
  resolution = "monthly"
))

#' @export
#' @rdname convenience
cdr_antarctic_daily <- make_cdr(list(
  hemisphere = "south",
  resolution = "daily"
))

#' @export
#' @rdname convenience
cdr_arctic_monthly <- make_cdr(list(
  hemisphere = "north",
  resolution = "monthly"
))

#' @export
#' @rdname convenience
cdr_arctic_daily <- make_cdr(list(hemisphere = "north", resolution = "daily"))

file_name <- function(url) {
  digest::digest(url)
}


cdr_variables_daily <- list(
  v4 = list(
    aice = "cdr_seaice_conc",
    qa = "qa_of_cdr_seaice_conc",
    stdev = "stdev_of_cdr_seaice_conc",
    interpolation_spatial = "spatial_interpolation_flag",
    interpolation_temporal = "temporal_interpolation_flag",
    aice_bt = "nsidc_bt_seaice_conc",
    aice_nt = "nsidc_nt_seaice_conc"
  ),
  v5 = list(
    aice = "cdr_seaice_conc",
    qa = "cdr_seaice_conc_qa_flag",
    stdev = "cdr_seaice_conc_stdev",
    interpolation_spatial = "cdr_seaice_conc_interp_spatial_flag",
    interpolation_temporal = "cdr_seaice_conc_interp_temporal_flag"
  ),
  v6 = list(
    aice = "cdr_seaice_conc",
    qa = "cdr_seaice_conc_qa_flag",
    stdev = "cdr_seaice_conc_stdev",
    interpolation_spatial = "cdr_seaice_conc_interp_spatial_flag",
    interpolation_temporal = "cdr_seaice_conc_interp_temporal_flag"
  )
)

cdr_variables_monthly <- list(
  v4 = list(
    aice = "cdr_seaice_conc_monthly",
    qa = "qa_of_cdr_seaice_conc_monthly",
    stdev = "stdev_of_cdr_seaice_conc_monthly",
    aice_bt = "nsidc_bt_seaice_conc_monthly",
    aice_nt = "nsidc_nt_seaice_conc_monthly"
  ),

  v5 = list(
    aice = "cdr_seaice_conc_monthly",
    stdev = "cdr_seaice_conc_monthly_stdev",
    qa = "cdr_seaice_conc_monthly_qa_flag"
  ),

  v6 = list(
    aice = "cdr_seaice_conc_monthly",
    stdev = "cdr_seaice_conc_monthly_stdev",
    qa = "cdr_seaice_conc_monthly_qa_flag"
  )
)

supported_versions <- c(4, 5, 6)

cdr_variables <- list(
  monthly = cdr_variables_monthly,
  daily = cdr_variables_daily
)

cdr_variables_description <- list(
  aice = "CDR Sea ice concentration.",
  qa = "Quality control flag.",
  stdev = "Sea ice standard deviation.",
  interpolation_spatial = "Flag for spatial interpolation.",
  interpolation_temporal = "Flag for temporal interpolation.",
  aice_bt = "Sea ice concentration of the Bootstrap method.",
  aice_nt = "Sea ice concentration of the Nasa Team method."
)

as_describe <- function(cdr_variables_description) {
  items <- paste0(
    "\\item{`",
    names(cdr_variables_description),
    "`}{",
    unlist(cdr_variables_description),
    "}"
  )

  paste0("\\describe{", paste0(items, collapse = "\n"), "}")
}


variable_name <- function(variables, version, resolution) {
  version <- paste0("v", version)

  cdr_variables[[resolution]][[version]][variables]
}


#' Download sea ice concentration from NSIDC Climate Data Record
#'
#' This is a low-level function to download data from PolarWatch's ERDDAP
#' server.
#'
#' @param date_range Vector of size two with the start and end dates. Supported
#' formats are:
#'   * A Date or POSIXct/POSIXlt object.
#'   * A character vector with ISO format dates: `c("2020-01-01", "2020-12-31")`
#'   * A character vector with year-month: `c("2020-01", "2020-06")` (expands to first/last day of month).
#'   * A character vector with year only: `c("2020", "2021")` (expands to full year).
#' @param variables Character vector with the variables to fetch. Valid values are
#' `r as_describe(cdr_variables_description)`
#' Although not all variables are available in all versions and all resolutions.
#' The `*_nt` and `*_bt` variables are only available in version 4 and the
#' interpolation flags are only available for daily values.
#' @param hemisphere Character with the hemisphere to download.
#' Can be either "south" or "north".
#' @param resolution Character with the temporal resolution.
#' Can be either "monthly" or "daily".
#' @param date_stride Numeric with the temporal stride.
#' A number greater than 1 means to take only the nth date in the series.
#' @param xgrid_range Numeric vector of size 2 with the range of the x dimension.
#' @param xgrid_stride Numeric with the stride of the x dimension.
#' @param ygrid_range Numeric vector of size 2 with the range of the y dimension.
#' @param ygrid_stride Numeric with the stride of the y dimension.
#' @param format Character with the format.
#' @param version Version of the dataset. Can be 4 or 5.
#' @param file Character with the file name to use for download. If `NULL`,
#' the file name will be constructed by hashing the request URL.
#' Requests consisting in more than one file will append a number.
#' @param dir Directory where to download the file.
#' @param use_cache Logical indicating whether to not perform the download if
#' the file already exist.
#'
#'
#' @returns Path or vector of paths to the downloaded files.
#'
#' @examples
#' \dontrun{
#' cdr(date_range = c("2022-01", "2022-01"),
#'     # Data every 7 days
#'     date_stride = 7,
#'     resolution = "daily",
#'     # Thin the grid by taking every other gridpoint
#'     xgrid_stride = 2,
#'     ygrid_stride = 2,
#'     hemisphere = "north"
#'     )
#'}
#' @export
cdr <- function(
  date_range = c(NA, NA),

  variables = "aice",
  hemisphere = c("south", "north"),
  resolution = c("monthly", "daily"),

  date_stride = 1,

  xgrid_range = c(NA, NA),
  xgrid_stride = 1,

  ygrid_range = c(NA, NA),
  ygrid_stride = 1,
  version = 6,
  format = "nc",
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
) {
  if (!(version %in% supported_versions)) {
    cli::cli_abort(
      "Invalid version. Available versions are {supported_versions}"
    )
  }

  resolution <- resolution[1]
  resolutions <- c("monthly", "daily")
  if (!(resolution %in% resolutions)) {
    cli::cli_abort(
      "{.arg resolutions} needs to be one of {.val {resolutions}}, not {.val {resolution}}."
    )
  }

  variables_translated <- variable_name(variables, version, resolution)

  missing_variable <- vapply(variables_translated, is.null, logical(1))
  if (any(missing_variable)) {
    bad_variables <- variables[missing_variable]
    cli::cli_abort(c(
      "Variable{?s} not available for this dataset: {.val {bad_variables}}.",
      i = "See possible variables in {.help [{.fun cdr}](icecdr::cdr)}."
    ))
  }

  hemisphere <- hemisphere[1]
  hemispheres <- c("south", "north")
  if (!(hemisphere %in% hemispheres)) {
    cli::cli_abort(
      "{.arg hemisphere} needs to be one of {.val {hemispheres}}, not {.val {hemisphere}}."
    )
  }

  date_range <- parse_range(date_range)

  if (resolution == "monthly") {
    date_range <- floor_month(date_range)
  }

  hemisphere <- hemisphere[1]
  if (hemisphere == "north") {
    ygrid_limits <- c(5837500.0, -5337500.0)
    xgrid_limits <- c(-3837500.0, 3737500.0)
  }

  if (hemisphere == "south") {
    ygrid_limits <- c(4337500.0, -3937500.0)
    xgrid_limits <- c(-3937500.0, 3937500.0)
  }

  xgrid_range <- replace_limits(xgrid_range, xgrid_limits)
  ygrid_range <- replace_limits(ygrid_range, ygrid_limits)

  # This came from testing the versions. V4 reports
  # half the size than version 5. Must be the level of
  # compression
  size_per_point <- if (version == 4) 4 else 8
  size <- size_per_point *
    request_size(
      variables,
      resolution,
      date_range,
      date_stride,
      xgrid_range,
      xgrid_stride,
      ygrid_range,
      ygrid_stride
    )

  # 2gb file limit
  size_limit <- 2 * 1024 * 1024 * 1024
  n_chunks <- ceiling(size / size_limit)

  if (n_chunks > 1) {
    class(size) <- "object_size"
    size <- format(size, "auto")
    cli::cli_inform(
      "Request size ({size}) larger than the limit (2 Gb). Downloading in {n_chunks} chunks."
    )

    if (resolution == "daily") {
      unit <- paste0(date_stride, " day")
    }

    if (resolution == "monthly") {
      unit <- paste0(date_stride, " month")
    }

    dates <- seq(date_range[1], date_range[2], by = unit)
    chunks <- as.numeric(cut(dates, n_chunks))

    files <- vapply(
      unique(chunks),
      function(chunk) {
        date_range <- range(dates[chunks == chunk])

        if (!is.null(file)) {
          file <- paste0(
            tools::file_path_sans_ext(file),
            "_",
            formatC(chunk, width = 2, flag = "0"),
            ".",
            tools::file_ext(file)
          )
        }

        cli::cli_inform(
          "Downloading file {chunk} of {n_chunks} ({date_range[1]} to {date_range[2]})."
        )

        cdr(
          date_range = date_range,
          variables = variables,
          hemisphere = hemisphere,
          resolution = resolution,
          date_stride = date_stride,
          xgrid_range = xgrid_range,
          xgrid_stride = xgrid_stride,
          ygrid_range = ygrid_range,
          ygrid_stride = ygrid_stride,
          format = format,
          version = version,
          use_cache = use_cache,
          dir = dir,
          file = file
        )
      },
      character(1)
    )

    return(files)
  }

  url <- nsidc_url(
    variables = variables_translated,
    hemisphere = hemisphere,
    resolution = resolution,
    date_range = date_range,
    date_stride = date_stride,
    xgrid_range = xgrid_range,
    xgrid_stride = xgrid_stride,
    ygrid_range = ygrid_range,
    ygrid_stride = ygrid_stride,
    version = version,
    format = format
  )

  if (is.null(file)) {
    file <- paste0(file_name(url), ".", format)
  }

  if (!dir.exists(dir)) {
    dir.create(dir)
  }

  destination <- file.path(dir, file)
  source_file <- paste0(destination, ".source")

  if (use_cache) {
    if (all(file.exists(c(source_file, destination)))) {
      existing_url <- readLines(source_file)
      if (url == existing_url) {
        cli::cli_inform("Returning existing file.")
        return(destination)
      }
    }
  }

  old <- options(timeout = 60 * 360)
  on.exit(options(old))

  if (debug_no_download()) {
    writeLines("", destination)
    return(destination)
  }

  cli::cli_inform("Downloading data.")

  download_file(url, destination)

  writeLines(url, source_file)

  return(destination)
}

icecdr_user_agent <- "icecdr (https://github.com/eliocamp/icecdr)"

with_no_download <- function(expr) {
  old <- getOption("CDR_DONT_DOWNLOAD", default = FALSE)
  on.exit(options(CDR_DONT_DOWNLOAD = old))
  options(CDR_DONT_DOWNLOAD = TRUE)
  return(expr)
}

debug_no_download <- function() {
  getOption("CDR_DONT_DOWNLOAD", default = FALSE)
}

error_info <- function(response) {
  message <- httr2::resp_body_string(response) |>
    strsplit(split = "\n") |>
    _[[1]]

  message <- message[grep(message, pattern = "message")] |>
    gsub(pattern = "    message=\"", replacement = "") |>
    gsub(pattern = "\";", replacement = "") |>
    gsub(pattern = r"(\\\")", replacement = "\"", perl = TRUE)

  list(status_code = response$status_code, message = message)
}


request_size <- function(
  variables,
  resolution,
  date_range,
  date_stride,
  xgrid_range,
  xgrid_stride,
  ygrid_range,
  ygrid_stride
) {
  if (resolution == "daily") {
    unit <- paste0(date_stride, " day")
  }

  if (resolution == "monthly") {
    unit <- paste0(date_stride, " month")
  }

  ndates <- length(seq(
    date_range[1],
    date_range[2],
    by = unit
  ))

  res <- 25000
  nx <- (abs(xgrid_range[2] - xgrid_range[1]) / xgrid_stride) / res + 1
  ny <- (abs(ygrid_range[2] - ygrid_range[1]) / ygrid_stride) / res + 1
  nvar <- length(variables)

  nx * ny * ndates * nvar
}

nsidc_variables <- c(
  "cdr_seaice_conc",
  "melt_onset_day_cdr_seaice_conc",
  "nsidc_bt_seaice_conc",
  "nsidc_nt_seaice_conc",
  "qa_of_cdr_seaice_conc",
  "spatial_interpolation_flag",
  "stdev_of_cdr_seaice_conc",
  "temporal_interpolation_flag"
)

# Will change to
erdap_server <- "https://coastwatch.pfeg.noaa.gov/"

nsidc_url <- function(
  variables = "cdr_seaice_conc",
  hemisphere = c("south", "north"),
  resolution = c("monthly", "daily"),

  date_range = c(NA, NA),
  date_stride = 1,

  xgrid_range = c(NA, NA),
  xgrid_stride = 1,

  ygrid_range = c(NA, NA),
  ygrid_stride = 1,
  version = 5,
  format = "nc"
) {
  resolution <- c(daily = "1day", monthly = "mday")[resolution]
  hemisphere <- c(south = "sh", north = "nh")[hemisphere]

  base <- glue::glue(
    "{erdap_server}/erddap/griddap/nsidcG02202v{version}{hemisphere}{resolution}.{format}"
  )
  date_range_str <- format(date_range, "%Y-%m-%dT00:00:00Z")
  time <- variable_definition(date_range_str, date_stride)
  xgrid <- variable_definition(xgrid_range, xgrid_stride)
  ygrid <- variable_definition(ygrid_range, ygrid_stride)

  query <- paste0(
    glue::glue("{variables}[{time}][{ygrid}][{xgrid}]"),
    collapse = ","
  )

  utils::URLencode(glue::glue("{base}?{query}"))
}


replace_limits <- function(range, limits) {
  ifelse(is.na(range), limits, range)
}

variable_definition <- function(range, stride) {
  glue::glue("({range[1]}):{stride}:({range[2]})")
}


floor_month <- function(date) {
  y <- format(date, "%Y")
  m <- format(date, "%m")
  as.Date(glue::glue("{y}-{m}-01"))
}

test_that("cdr_returns files", {
  range <- c("2023-01-01", "2023-05-31")
  file <- with_no_download(cdr_antarctic_monthly(range, use_cache = FALSE))

  expect_length(file, 1)
  expect_type(file, "character")
  expect_true(file.exists(file))
})


test_that("long requets are splitted", {
  range <- c("1990-01-01", "2023-05-31")
  suppressMessages(expect_message(
    file <- with_no_download(cdr_arctic_daily(
      range,
      file = "arctic.nc",
      use_cache = FALSE
    ))
  ))

  expect_length(file, 7)
  expect_type(file, "character")

  expected <- vapply(
    1:7,
    function(x) paste0("arctic_0", x, ".nc"),
    character(1)
  )
  expect_equal(basename(file), expected)
})


test_that("cdr errors with bad range", {
  range <- c("1900-01", "2023-01")
  vcr::local_cassette("bad-range")
  expect_error(cdr_antarctic_monthly(range), "Not Found")
})


test_that("cache works", {
  range <- c("2023-01-01", "2023-01")
  suppressMessages(
    file <- with_no_download(cdr_antarctic_monthly(range, use_cache = TRUE))
  )
  info <- file.info(file)
  file2 <- with_no_download(cdr_antarctic_monthly(range, use_cache = TRUE))

  expect_equal(info, file.info(file2))
})


test_that("Error messages", {
  range <- c("2023-01", "2023-01")
  expect_error(
    cdr_antarctic_monthly(
      variables = "random variable",
      range
    ),
    "Variable not available"
  )

  expect_error(cdr(date_range = range, hemisphere = "antarctica"), "hemisphere")

  expect_error(cdr(date_range = range, hemisphere = "antarctica"), "hemisphere")

  expect_error(cdr(date_range = range, resolution = "yearly"), "resolution")
})

zero_pad <- function(x, width = 2) {
  formatC(x, width = width, flag = "0")
}

download_files <- function(urls, destinations = NULL) {
  if (is.null(destinations)) {
    destinations <- replicate(length(urls), tempfile())
  }

  return(invisible(unname(mapply(download_file, urls, destinations))))
}

download_file <- function(url, destination = NULL) {
  if (is.null(destination)) {
    destination <- tempfile()
  }

  response <- httr2::request(url) |>
    httr2::req_user_agent(icecdr_user_agent) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_progress(type = "down") |>
    httr2::req_perform(path = destination)

  # When mocking requests for tests, the response is not saved to file, so
  # we need to handle that ourselves.
  if (is.raw(response$body)) {
    writeBin(response$body, destination)
  }

  if (httr2::resp_is_error(response)) {
    info <- error_info(response)
    if (file.exists(destination)) {
      file.remove(destination)
    }
    error_msg <- "Failed to download. Got error {info$status_code}"

    if (length(info$message) != 0) {
      error_msg <- c(paste0(error_msg, " with message:"), info$message)
    } else {
      error_msg <- paste0(error_msg, ".")
    }
    cli::cli_abort(error_msg)
  }
  return(invisible(destination))
}

parse_range <- function(x) {
  if (is_dateish(x)) {
    return(x)
  }

  x <- strsplit(as.character(x), "-")
  x <- lapply(x, as.numeric)

  # Support for inputting year
  if (length(x[[1]]) == 1) {
    x[[1]][[2]] <- 1
    x[[1]][[3]] <- 1
  }

  if (length(x[[2]]) == 1) {
    x[[2]][[2]] <- 12
    x[[2]][[3]] <- 31
  }

  # Support for year-month
  if (length(x[[1]]) == 2) {
    x[[1]][[3]] <- 1
  }

  if (length(x[[2]]) == 2) {
    # Need to convert the month into date to correctly
    # account for leap years
    days <- days_in_month(as.Date(paste0(c(x[[2]], 1), collapse = "-")))

    x[[2]][[3]] <- days
  }


  as.Date(vapply(x, function(x) paste0(x, collapse = "-"), character(1)))
}



is_dateish <- function(x) {
  inherits(x, "Date") | inherits(x, "POSIXct") | inherits(x, "POSIXlt")
}


# from lubridate
leap_year <- function(date) {
  if (is.numeric(date)) {
    year <- date
  } else {
    year <- year(date)
  }
  (year %% 4 == 0) & ((year %% 100 != 0) | (year %% 400 == 0))
}

N_DAYS_IN_MONTHS <- c(
  Jan = 31L,
  Feb = 28L,
  Mar = 31L,
  Apr = 30L,
  May = 31L,
  Jun = 30L,
  Jul = 31L,
  Aug = 31L,
  Sep = 30L,
  Oct = 31L,
  Nov = 30L,
  Dec = 31L
)

year <- function(date) {
  as.numeric(format(date, "%Y"))
}

month <- function(date) {
  as.numeric(format(date, "%m"))
}

days_in_month <- function(x) {
  month_x <- month(x)
  n_days <- N_DAYS_IN_MONTHS[month_x]
  n_days[month_x == 2 & leap_year(x)] <- 29L
  n_days
}

test_that("character and date range works", {
  range <- c("2015-01-15", "2015-03-13")
  expect_equal(parse_range(range), as.Date(range))

  range <- as.Date(c("2015-01-15", "2015-03-13"))
  expect_equal(parse_range(range), range)
})


test_that("year-month range works", {
  range <- c("2015-01", "2015-03-13")
  expect_equal(parse_range(range), as.Date(c("2015-01-01", range[2])))

  range <- c("2015-01-03", "2015-03")
  expect_equal(parse_range(range), as.Date(c("2015-01-03", "2015-03-31")))
})


test_that("year works", {
  range <- c("2015", "2015-03-13")
  expect_equal(parse_range(range), as.Date(c("2015-01-01", range[2])))

  range <- c("2015", "2015")
  expect_equal(parse_range(range), as.Date(c("2015-01-01", "2015-12-31")))

})

test_that("leap years work", {
  range <- c("2000", "2004-02")
  expect_equal(parse_range(range), as.Date(c("2000-01-01", "2004-02-29")))

  # Skip years (not really relevant to this package, though, but still nice)
  range <- c("2000", "1900-02")
  expect_equal(parse_range(range), as.Date(c("2000-01-01", "1900-02-28")))
})

#' Download sea ice index data.
#'
#' The NSIDC sea ice index product provides computed sea ice extent for each hemisphere
#' on daily and monthly resolution.
#'
#' @inheritParams cdr
#'
#' @details
#' The returned files are not the raw files, but contain minor format modifications
#' for ease of use. Instead of `year` and `month` columns, there is a single `time`
#' column.
#' Instead of a `region` column with `"S"` or `"N"` values, there is a `hemisphere`
#' column with values `"south"` and `"north"` (for consistency with the function
#' arguments).
#' Column names are all in lower case and are the same for the daily and monthly
#' products, except for an `area` column that is only available in the monthly product.
#' Area and extent values are expressed in km^2.
#'
#' @returns The route to the file name.
#'
#' @examples
#' \dontrun{
#' sea_ice_index("south", "monthly")
#' }
#'
#' @export
sea_ice_index <- function(
  hemisphere = c("south", "north"),
  resolution = c("monthly", "daily"),
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
) {
  resolution <- resolution[1]
  resolutions <- c("monthly", "daily")
  if (!(resolution %in% resolutions)) {
    cli::cli_abort(
      "{.arg resolutions} needs to be one of {.val {resolutions}}, not {.val {resolution}}."
    )
  }

  hemisphere <- hemisphere[1]
  hemispheres <- c("south", "north")
  if (!(hemisphere %in% hemispheres)) {
    cli::cli_abort(
      "{.arg hemisphere} needs to be one of {.val {hemispheres}}, not {.val {hemisphere}}."
    )
  }

  source <- glue::glue("{hemisphere}_{resolution}")
  if (is.null(file)) {
    file <- glue::glue("sea-ice-index_{source}.csv")
  }

  if (!dir.exists(dir)) {
    dir.create(dir)
  }

  destination <- file.path(dir, file)
  source_file <- paste0(destination, ".source")

  if (use_cache) {
    if (all(file.exists(c(source_file, destination)))) {
      existing_source <- readLines(source_file)
      if (source == existing_source) {
        cli::cli_inform("Returning existing file.")
        return(destination)
      }
    }
  }

  h <- if (hemisphere == "south") "S" else "N"

  cli::cli_inform("Downloading data.")

  if (resolution == "daily") {
    url <- glue::glue(
      "https://noaadata.apps.nsidc.org/NOAA/G02135/{hemisphere}/daily/data/{h}_seaice_extent_daily_v4.0.csv"
    )

    destination <- download_files(url, destination)
    destination <- fix_daily(destination, hemisphere)
  } else {
    m <- zero_pad(1:12)
    files <- glue::glue("{h}_{m}_extent_v4.0.csv")
    urls <- glue::glue(
      "https://noaadata.apps.nsidc.org/NOAA/G02135/{hemisphere}/monthly/data/{files}"
    )

    tempfiles <- download_files(urls)
    destination <- merge_monthly_files(tempfiles, destination)
  }

  writeLines(source, source_file)

  return(destination)
}

fix_daily <- function(file, hemisphere) {
  data <- utils::read.csv(file, header = FALSE, skip = 2)

  data$time <- as.POSIXct(
    as.Date(
      glue::glue("{data$V1}-{data$V2}-{data$V3}")
    ),
    tz = "utc"
  )
  data$extent <- data$V4
  data$hemisphere <- hemisphere
  data$extent <- data$extent * 1e12

  data <- data[, c("time", "hemisphere", "extent")]
  utils::write.csv(data, file, row.names = FALSE)
  return(file)
}

merge_monthly_files <- function(in_files, out_file) {
  data <- lapply(in_files, utils::read.csv)
  data <- Reduce(rbind, data)

  data$time <- as.POSIXct(
    as.Date(
      glue::glue("{data$year}-{data$mo}-01")
    ),
    tz = "utc"
  )
  data$hemisphere <- ifelse(trimws(data$region) == "S", "south", "north")
  data$extent <- data$extent * 1e12
  data$area <- data$area * 1e12

  data <- data[, c("time", "hemisphere", "extent", "area")]
  utils::write.csv(data, out_file, row.names = FALSE)
  return(out_file)
}

test_that("monthly sea_ice_index gets data", {
  vcr::local_cassette("sea_index_monthly")
  file <- sea_ice_index(
    hemisphere = "south",
    resolution = "monthly",
    use_cache = FALSE
  )

  expect_true(file.exists(file))

  data <- read.csv(file)
  expect_true(nrow(data) > 0)
  expect_identical(colnames(data), c("time", "hemisphere", "extent", "area"))
  expect_true(all(data[["hemisphere"]] == "south"))
})

test_that("monthly sea_ice_index gets data", {
  vcr::local_cassette("sea_index_daily")
  file <- sea_ice_index(
    hemisphere = "north",
    resolution = "daily",
    use_cache = FALSE
  )

  expect_true(file.exists(file))

  data <- read.csv(file)
  expect_true(nrow(data) > 0)
  expect_identical(colnames(data), c("time", "hemisphere", "extent"))
  expect_true(all(data[["hemisphere"]] == "north"))
})


test_that("sea_ice_index throws good errors", {
  expect_error(sea_ice_index(hemisphere = "both"), "hemisphere")
  expect_error(sea_ice_index(resolution = "both"), "resolution")
})


test_that("sea_ice_index cache works", {
  vcr::local_cassette("sea_index_monthly")
  dir <- tempfile()
  expect_message(
    file <- sea_ice_index(dir = dir, use_cache = TRUE),
    "Downloading"
  )

  info <- file.info(file)
  expect_message(
    file2 <- sea_ice_index(dir = dir, use_cache = TRUE),
    "existing"
  )

  expect_equal(info, file.info(file2))
})

#' Minor fixes to CDR files.
#'
#' `cdr_fix()` "fixes" some inconsistencies in the raw data.
#' It standardised variable names between versions and temporal resolutions and
#' adds projection information to the grid definition so CDO can compute the
#' area of each gridpoint to compute weighted means or  area integrals.
#' Both require the `rcdo` package and CDO installed.
#'
#' @param files Path to the files.
#' @param fix Character vector with the fixes to implement.
#'
#' @details
#' `cdr_fix_names()` and `cdr_fix_grid()` are aliases for `cdr_fix(fix = "names")`
#' and `cdr_fix(fix = "grid")`, respectively.
#' Using `cdr_fix()` to fix both at the same time is more efficient as it
#' recreates each file only once.
#'
#' @return The path to the modified files.
#'
#' @examples
#' \dontrun{
#' cdr_antarctic_monthly(c("2022-01", "2022-01")) |>
#'   cdr_fix()
#' }
#'
#' @export
#' @rdname fixes
cdr_fix <- function(files, fix = c("names", "grid")) {
  vapply(files, cdr_fix_one, fix = fix, FUN.VALUE = character(1))
}

#' @export
#' @rdname fixes
cdr_fix_names <- function(files) {
  cdr_fix(files, fix = "names")
}

#' @export
#' @rdname fixes
cdr_fix_grid <- function(files) {
  cdr_fix(files, fix = "grid")
}


cdr_fix_one <- function(file, fix = c("names", "grid")) {
  if (!file.exists(file)) {
    cli::cli_abort("File {.file {file}} does not exist.")
  }

  rlang::check_installed(c("rcdo", "ncdf4"))

  op <- file

  if ("grid" %in% fix) {
    op <- cdo_fix_grid(file, op)
  }

  if ("names" %in% fix) {
    op <- cdo_fix_name(file, op)
  }

  if (is.character(op)) {
    return(op)
  }

  out <- rcdo::cdo_execute(op)
  file_move(out, file)

  return(file)
}


cdo_fix_grid <- function(file, prev_op = file) {
  grid <- rcdo::cdo_execute(rcdo::cdo_griddes(file))

  grid <- grid_parse(grid)

  if (grid_is_valid(grid)) {
    return(prev_op)
  }

  grid$gridtype <- "projection"

  fnc <- ncdf4::nc_open(file)
  on.exit(ncdf4::nc_close(fnc))

  # Remove possible old params
  grid <- grid[!grepl("grid_mapping", grid)]
  grid <- grid[!grepl("grid_mapping_name", grid)]
  grid <- grid[!grepl("proj_params", grid)]
  grid <- grid[!grepl("scanningMode", grid)]

  proj4 <- ncdf4::ncatt_get(fnc, 0, "grid_mapping_proj4text")

  if (!proj4$hasatt) {
    cli::cli_abort("no crs")
  }

  grid_mapping_name <- ncdf4::ncatt_get(fnc, 0, "grid_mapping_name")$value

  grid$grid_mapping <- "crs"
  grid$grid_mapping_name <- grid_mapping_name
  grid$proj_params <- glue::glue("\"{proj4$value}\"")

  grid_file <- grid_write(grid, tempfile())

  op <- rcdo::cdo_setgrid(prev_op, grid_file)

  return(op)
}

grid_parse <- function(grid) {
  grid <- grid[!grepl("^#", grid)]

  splitted <- strsplit(grid, split = "=")

  names <- vapply(splitted, function(x) trimws(x[[1]]), character(1))
  values <- lapply(splitted, function(x) trimws(paste0(x[-1], collapse = "=")))

  stats::setNames(values, names)
}

grid_write <- function(grid, file) {
  text <- paste0(names(grid), " = ", unlist(grid, recursive = FALSE))
  writeLines(text, file)
  return(invisible(file))
}

grid_is_valid <- function(grid) {
  grid$gridtype == "projection" &
    !is.null(grid$grid_mapping) &
    !is.null(grid$grid_mapping_name) &
    !is.null(grid$proj_params)
}


cdo_fix_name <- function(file, prev_op = file) {
  old_variables <- strsplit(
    trimws(rcdo::cdo_execute(rcdo::cdo_showname(file))),
    split = " "
  )[[1]]

  vars <- unlist(cdr_variables)
  vars <- vars[!duplicated(vars)] # unique(x) removes names :(

  vars <- vars[vars %in% old_variables]

  if (length(vars) == 0) {
    cli::cli_inform(
      "No standard variable name found, returning unchanged file"
    )
    return(prev_op)
  }

  new_variables <- vapply(
    strsplit(names(vars), split = ".", fixed = TRUE),
    function(x) x[3],
    character(1)
  )

  names <- paste0(
    paste(old_variables, new_variables, sep = ","),
    collapse = ","
  )
  op <- rcdo::cdo_chname(prev_op, names)
  return(op)
}


# We can't really "move" if from and to
# are on different drives.
# Solution from https://github.com/bluegreen-labs/ecmwfr/blob/365532ef763792bdfb788616f3a383c97f3fb9a6/R/service-ds.R#L230-L240
file_move <- function(from, to) {
  # Copy data from temporary file to final location
  move <- suppressWarnings(file.rename(from, to))

  if (!move) {
    file.copy(from, to, overwrite = TRUE)
    file.remove(from)
  }
  return(to)
}

cdo_installed <- system("cdo -V", ignore.stdout = TRUE) == 0
skip_if(!cdo_installed)

test_that("fix all works", {
  file_og <- "test_data/6d5fafd0d3fa213a0a0943dc0d87ec8e.nc"
  file <- tempfile()
  file.copy(file_og, file)

  file_is_valid_grid <- function(file) {
    grid_is_valid(grid_parse(rcdo::cdo_execute(rcdo::cdo_griddes(file))))
  }

  file <- suppressWarnings(cdr_fix(file))

  expect_true(file_is_valid_grid(file))

  get_vars <- function(file) {
    trimws(rcdo::cdo_execute(rcdo::cdo_showname(file)))
  }

  expect_equal(
    get_vars(file),
    "aice"
  )
})


test_that("fix grid works", {
  file_og <- "test_data/6d5fafd0d3fa213a0a0943dc0d87ec8e.nc"
  file <- tempfile()
  file.copy(file_og, file)

  file_is_valid_grid <- function(file) {
    grid_is_valid(grid_parse(rcdo::cdo_execute(rcdo::cdo_griddes(file))))
  }

  file <- suppressWarnings(cdr_fix_grid(file))

  expect_true(file_is_valid_grid(file))
})


test_that("fix names works", {
  file_og <- "test_data/2f607d15cbb6ed73b98555c196e9eee1.nc"

  file <- tempfile()
  file.copy(file_og, file)

  file <- cdr_fix_names(file)

  get_vars <- function(file) {
    trimws(rcdo::cdo_execute(rcdo::cdo_showname(file)))
  }

  expect_equal(
    get_vars(file),
    "aice aice_bt"
  )

  files <- replicate(2, {
    file <- tempfile()
    file.copy(file_og, file)
    file
  })

  files <- cdr_fix_names(files)

  expect_equal(
    get_vars(files[1]),
    "aice aice_bt"
  )

  expect_equal(
    get_vars(files[2]),
    "aice aice_bt"
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cdr-fix.R
\name{cdr_fix}
\alias{cdr_fix}
\alias{cdr_fix_names}
\alias{cdr_fix_grid}
\title{Minor fixes to CDR files.}
\usage{
cdr_fix(files, fix = c("names", "grid"))

cdr_fix_names(files)

cdr_fix_grid(files)
}
\arguments{
\item{files}{Path to the files.}

\item{fix}{Character vector with the fixes to implement.}
}
\value{
The path to the modified files.
}
\description{
\code{cdr_fix()} "fixes" some inconsistencies in the raw data.
It standardised variable names between versions and temporal resolutions and
adds projection information to the grid definition so CDO can compute the
area of each gridpoint to compute weighted means or  area integrals.
Both require the \code{rcdo} package and CDO installed.
}
\details{
\code{cdr_fix_names()} and \code{cdr_fix_grid()} are aliases for \code{cdr_fix(fix = "names")}
and \code{cdr_fix(fix = "grid")}, respectively.
Using \code{cdr_fix()} to fix both at the same time is more efficient as it
recreates each file only once.
}
\examples{
\dontrun{
cdr_antarctic_monthly(c("2022-01", "2022-01")) |>
  cdr_fix()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cdr.R
\name{cdr}
\alias{cdr}
\title{Download sea ice concentration from NSIDC Climate Data Record}
\usage{
cdr(
  date_range = c(NA, NA),
  variables = "aice",
  hemisphere = c("south", "north"),
  resolution = c("monthly", "daily"),
  date_stride = 1,
  xgrid_range = c(NA, NA),
  xgrid_stride = 1,
  ygrid_range = c(NA, NA),
  ygrid_stride = 1,
  version = 6,
  format = "nc",
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)
}
\arguments{
\item{date_range}{Vector of size two with the start and end dates. Supported
formats are:
\itemize{
\item A Date or POSIXct/POSIXlt object.
\item A character vector with ISO format dates: \code{c("2020-01-01", "2020-12-31")}
\item A character vector with year-month: \code{c("2020-01", "2020-06")} (expands to first/last day of month).
\item A character vector with year only: \code{c("2020", "2021")} (expands to full year).
}}

\item{variables}{Character vector with the variables to fetch. Valid values are
\describe{\item{\code{aice}}{CDR Sea ice concentration.}
\item{\code{qa}}{Quality control flag.}
\item{\code{stdev}}{Sea ice standard deviation.}
\item{\code{interpolation_spatial}}{Flag for spatial interpolation.}
\item{\code{interpolation_temporal}}{Flag for temporal interpolation.}
\item{\code{aice_bt}}{Sea ice concentration of the Bootstrap method.}
\item{\code{aice_nt}}{Sea ice concentration of the Nasa Team method.}}
Although not all variables are available in all versions and all resolutions.
The \verb{*_nt} and \verb{*_bt} variables are only available in version 4 and the
interpolation flags are only available for daily values.}

\item{hemisphere}{Character with the hemisphere to download.
Can be either "south" or "north".}

\item{resolution}{Character with the temporal resolution.
Can be either "monthly" or "daily".}

\item{date_stride}{Numeric with the temporal stride.
A number greater than 1 means to take only the nth date in the series.}

\item{xgrid_range}{Numeric vector of size 2 with the range of the x dimension.}

\item{xgrid_stride}{Numeric with the stride of the x dimension.}

\item{ygrid_range}{Numeric vector of size 2 with the range of the y dimension.}

\item{ygrid_stride}{Numeric with the stride of the y dimension.}

\item{version}{Version of the dataset. Can be 4 or 5.}

\item{format}{Character with the format.}

\item{file}{Character with the file name to use for download. If \code{NULL},
the file name will be constructed by hashing the request URL.
Requests consisting in more than one file will append a number.}

\item{dir}{Directory where to download the file.}

\item{use_cache}{Logical indicating whether to not perform the download if
the file already exist.}
}
\value{
Path or vector of paths to the downloaded files.
}
\description{
This is a low-level function to download data from PolarWatch's ERDDAP
server.
}
\examples{
\dontrun{
cdr(date_range = c("2022-01", "2022-01"),
    # Data every 7 days
    date_stride = 7,
    resolution = "daily",
    # Thin the grid by taking every other gridpoint
    xgrid_stride = 2,
    ygrid_stride = 2,
    hemisphere = "north"
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sea_ice_index.R
\name{sea_ice_index}
\alias{sea_ice_index}
\title{Download sea ice index data.}
\usage{
sea_ice_index(
  hemisphere = c("south", "north"),
  resolution = c("monthly", "daily"),
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)
}
\arguments{
\item{hemisphere}{Character with the hemisphere to download.
Can be either "south" or "north".}

\item{resolution}{Character with the temporal resolution.
Can be either "monthly" or "daily".}

\item{file}{Character with the file name to use for download. If \code{NULL},
the file name will be constructed by hashing the request URL.
Requests consisting in more than one file will append a number.}

\item{dir}{Directory where to download the file.}

\item{use_cache}{Logical indicating whether to not perform the download if
the file already exist.}
}
\value{
The route to the file name.
}
\description{
The NSIDC sea ice index product provides computed sea ice extent for each hemisphere
on daily and monthly resolution.
}
\details{
The returned files are not the raw files, but contain minor format modifications
for ease of use. Instead of \code{year} and \code{month} columns, there is a single \code{time}
column.
Instead of a \code{region} column with \code{"S"} or \code{"N"} values, there is a \code{hemisphere}
column with values \code{"south"} and \code{"north"} (for consistency with the function
arguments).
Column names are all in lower case and are the same for the daily and monthly
products, except for an \code{area} column that is only available in the monthly product.
Area and extent values are expressed in km^2.
}
\examples{
\dontrun{
sea_ice_index("south", "monthly")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/icecdr-package.R
\docType{package}
\name{icecdr-package}
\alias{icecdr}
\alias{icecdr-package}
\title{icecdr: Download Sea Ice Concentration Data from the NSIDC Climate Data Record}
\description{
Programmatic access to NSIDC's sea ice concentration CDR \url{https://nsidc.org/data/g02202} via ERDAPP server and Sea Ice index \url{https://nsidc.org/data/g02135}. Supports caching results and optional fixes for some inconsistencies of the raw files.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/eliocamp/icecdr}
  \item \url{https://eliocamp.github.io/icecdr/}
  \item Report bugs at \url{https://github.com/eliocamp/icecdr/issues}
}

}
\author{
\strong{Maintainer}: Elio Campitelli \email{eliocampitelli@gmail.com} (\href{https://orcid.org/0000-0002-7742-9230}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Elio Campitelli \email{eliocampitelli@gmail.com} (\href{https://orcid.org/0000-0002-7742-9230}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cdr.R
\name{cdr_antarctic_monthly}
\alias{cdr_antarctic_monthly}
\alias{cdr_antarctic_daily}
\alias{cdr_arctic_monthly}
\alias{cdr_arctic_daily}
\title{Convenience functions}
\usage{
cdr_antarctic_monthly(
  date_range,
  variables = "aice",
  version = 6,
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)

cdr_antarctic_daily(
  date_range,
  variables = "aice",
  version = 6,
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)

cdr_arctic_monthly(
  date_range,
  variables = "aice",
  version = 6,
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)

cdr_arctic_daily(
  date_range,
  variables = "aice",
  version = 6,
  file = NULL,
  dir = tempdir(),
  use_cache = TRUE
)
}
\arguments{
\item{date_range}{Vector of size two with the start and end dates. Supported
formats are:
\itemize{
\item A Date or POSIXct/POSIXlt object.
\item A character vector with ISO format dates: \code{c("2020-01-01", "2020-12-31")}
\item A character vector with year-month: \code{c("2020-01", "2020-06")} (expands to first/last day of month).
\item A character vector with year only: \code{c("2020", "2021")} (expands to full year).
}}

\item{variables}{Character vector with the variables to fetch. Valid values are
\describe{\item{\code{aice}}{CDR Sea ice concentration.}
\item{\code{qa}}{Quality control flag.}
\item{\code{stdev}}{Sea ice standard deviation.}
\item{\code{interpolation_spatial}}{Flag for spatial interpolation.}
\item{\code{interpolation_temporal}}{Flag for temporal interpolation.}
\item{\code{aice_bt}}{Sea ice concentration of the Bootstrap method.}
\item{\code{aice_nt}}{Sea ice concentration of the Nasa Team method.}}
Although not all variables are available in all versions and all resolutions.
The \verb{*_nt} and \verb{*_bt} variables are only available in version 4 and the
interpolation flags are only available for daily values.}

\item{version}{Version of the dataset. Can be 4 or 5.}

\item{file}{Character with the file name to use for download. If \code{NULL},
the file name will be constructed by hashing the request URL.
Requests consisting in more than one file will append a number.}

\item{dir}{Directory where to download the file.}

\item{use_cache}{Logical indicating whether to not perform the download if
the file already exist.}
}
\value{
A string vector with the location of the downloaded files
}
\description{
These are convenience functions to download, daily or monthly
sea ice concentration data from the whole Antarctic or Arctic.
For a more complete low level function, see \code{\link[=cdr]{cdr()}}.
}
\examples{
\dontrun{
cdr_antarctic_monthly(c("2022-01", "2022-01"))
}


}
